#include "als_ltr_303als.h"

#define ALS_REG_CONTR             0x80U
#define ALS_REG_MEAS_RATE         0x85U
#define ALS_REG_PART_ID           0x86U
#define ALS_REG_MANUFAC_ID        0x87U
#define ALS_REG_DATA_CH1_0        0x88U
#define ALS_REG_DATA_CH1_1        0x89U
#define ALS_REG_DATA_CH0_0        0x8AU
#define ALS_REG_DATA_CH0_1        0x8BU
#define ALS_REG_STATUS            0x8CU
#define ALS_REG_INTERRUPT         0x8FU
#define ALS_REG_THRES_UP_0        0x97U
#define ALS_REG_THRES_UP_1        0x98U
#define ALS_REG_THRES_LOW_0       0x99U
#define ALS_REG_THRES_LOW_1       0x9AU
#define ALS_REG_INTERRUPT_PERSIST 0x9EU

#define ALS_REG_CONTR_MODE_Pos (0U)
#define ALS_REG_CONTR_MODE_Msk (1U << ALS_REG_CONTR_MODE_Pos)

#define ALS_REG_CONTR_SW_RST_Pos (1U)
#define ALS_REG_CONTR_SW_RST_Msk (1U << ALS_REG_CONTR_SW_RST_Pos)

#define ALS_REG_CONTR_GAIN_Pos (2U)
#define ALS_REG_CONTR_GAIN_Msk (7U << ALS_REG_CONTR_GAIN_Pos)

#define ALS_REG_MEAS_RATE_Pos (0U)
#define ALS_REG_MEAS_RATE_Msk (7U << ALS_REG_MEAS_RATE_Pos)

#define ALS_REG_MEAS_INT_Pos (3U)
#define ALS_REG_MEAS_INT_Msk (7U << ALS_REG_MEAS_INT_Pos)

#define ALS_REG_STATUS_DATA_VALID_Pos (7U)
#define ALS_REG_STATUS_DATA_VALID_Msk (1U << ALS_REG_STATUS_DATA_VALID_Pos)

#define ALS_REG_STATUS_GAIN_Pos (4U)
#define ALS_REG_STATUS_GAIN_Msk (7U << ALS_REG_STATUS_GAIN_Pos)

#define ALS_REG_STATUS_INT_Pos (3U)
#define ALS_REG_STATUS_INT_Msk (1U << ALS_REG_STATUS_INT_Pos)

#define ALS_REG_STATUS_DATA_Pos (2U)
#define ALS_REG_STATUS_DATA_Msk (1U << ALS_REG_STATUS_DATA_Pos)

#define ALS_REG_INTERRUPT_MODE_Pos (0U)
#define ALS_REG_INTERRUPT_MODE_Msk (1U << ALS_REG_INTERRUPT_MODE_Pos)

#define ALS_REG_INTERRUPT_POL_Pos (1U)
#define ALS_REG_INTERRUPT_POL_Msk (1U << ALS_REG_INTERRUPT_POL_Pos)

#define ALS_CONST_MFG_ID  0x05U
#define ALS_CONST_PART_ID 0x0AU

typedef enum {
    IMS_ALS_THRESH_LOW  = ALS_REG_THRES_LOW_0,
    IMS_ALS_THRESH_HIGH = ALS_REG_THRES_UP_0,
} ims_als_thresh_t;

typedef struct {
    uint8_t mfg_id;
    uint8_t part_id;
} ims_als_id_t;

static ims_ret_t ims_als_ltr_303als_reset(ims_ltr_303als_t *als);
static ims_ret_t ims_als_ltr_303als_read_id(ims_ltr_303als_t *als, ims_als_id_t *id);
static ims_ret_t ims_als_ltr_303als_set_gain(ims_ltr_303als_t *als, ims_ltr_303als_gain_t gain);
static ims_ret_t ims_als_ltr_303als_set_threshold(ims_ltr_303als_t *als, ims_als_thresh_t type, uint16_t thresh);
static ims_ret_t ims_ltr_303als_config_irq(ims_ltr_303als_t *als, ims_ltr_303als_int_polarity_t pol, bool enable);
static ims_ret_t ims_ltr_303als_config_timing(ims_ltr_303als_t *als, ims_ltr_303als_measurement_rate_t rate,
                                              ims_ltr_303als_integration_time_t time);

static ims_ret_t ims_als_ltr_303als_read_register(ims_ltr_303als_t *als, uint8_t reg, uint8_t *value);
static ims_ret_t ims_als_ltr_303als_write_register(ims_ltr_303als_t *als, uint8_t reg, uint8_t value);

ims_ret_t ims_ltr_303als_init(ims_ltr_303als_t *als) {
    ims_als_id_t id;

    if (ims_als_ltr_303als_reset(als) != IMS_SUCCESS) {
        return IMS_FAIL;
    }

    /* Read Mfg. ID and Part ID */
    if (ims_als_ltr_303als_read_id(als, &id) != IMS_SUCCESS) {
        return IMS_FAIL;
    }

    /* Validate IDs */
    if (id.mfg_id != ALS_CONST_MFG_ID || ((id.part_id & 0xF0U) >> 4U) != ALS_CONST_PART_ID) {
        return IMS_FAIL;
    }

    return IMS_SUCCESS;
}

void ims_ltr_303als_preset_config(ims_ltr_303als_config_t *cfg) {
    cfg->gain               = IMS_LTR_303ALS_GAIN_1X;
    cfg->interrupt_en       = true;
    cfg->interrupt_polarity = IMS_LTR_303ALS_INT_ACTIVE_LOW;
    cfg->threshold_low      = 0x0000U;
    cfg->threshold_high     = 0xFFFFU;
    cfg->integration_time   = IMS_LTR_303ALS_INT_250MS;
    cfg->measurement_rate   = IMS_LTR_303ALS_MEAS_500MS;
}

ims_ret_t ims_ltr_303als_config(ims_ltr_303als_t *als, ims_ltr_303als_config_t *config) {
    ims_ret_t ret;

    ret = ims_ltr_303als_config_timing(als, config->measurement_rate, config->integration_time);
    if (ret != IMS_SUCCESS) return ret;

    ret = ims_als_ltr_303als_set_gain(als, config->gain);
    if (ret != IMS_SUCCESS) return ret;

    ret = ims_als_ltr_303als_set_threshold(als, IMS_ALS_THRESH_LOW, config->threshold_low);
    if (ret != IMS_SUCCESS) return ret;

    ret = ims_als_ltr_303als_set_threshold(als, IMS_ALS_THRESH_HIGH, config->threshold_high);
    if (ret != IMS_SUCCESS) return ret;

    ret = ims_ltr_303als_config_irq(als, config->interrupt_polarity, config->interrupt_en);

    return ret;
}

ims_ret_t ims_ltr_303als_set_mode(ims_ltr_303als_t *als, ims_ltr_303als_mode_t mode) {
    ims_ret_t ret;

    uint8_t r_cont;

    ret = ims_als_ltr_303als_read_register(als, ALS_REG_CONTR, &r_cont);
    if (ret != IMS_SUCCESS) {
        return ret;
    }

    r_cont &= ~ALS_REG_CONTR_MODE_Msk;
    r_cont |= ((mode << ALS_REG_CONTR_MODE_Pos) & ALS_REG_CONTR_MODE_Msk);

    ret = ims_als_ltr_303als_write_register(als, ALS_REG_CONTR, r_cont);

    return ret;
}

ims_ret_t ims_ltr_303als_get_mode(ims_ltr_303als_t *als, ims_ltr_303als_mode_t *mode) {
    ims_ret_t ret;

    uint8_t r_cont;

    ret = ims_als_ltr_303als_read_register(als, ALS_REG_CONTR, &r_cont);
    if (ret != IMS_SUCCESS) {
        return ret;
    }

    if (r_cont & ALS_REG_CONTR_MODE_Msk) {
        *mode = IMS_LTR_303ALS_MODE_ACTIVE;
    } else {
        *mode = IMS_LTR_303ALS_MODE_STANDBY;
    }

    return IMS_SUCCESS;
}

ims_ret_t ims_ltr_303als_read(ims_ltr_303als_t *als, ims_ltr_303als_result_t *result) {
    ims_ret_t ret;

    uint8_t r_status;

    ret = ims_als_ltr_303als_read_register(als, ALS_REG_STATUS, &r_status);
    if (ret != IMS_SUCCESS) {
        return ret;
    }

    result->gain = (r_status & ALS_REG_STATUS_GAIN_Msk) >> ALS_REG_STATUS_GAIN_Pos;

    if (r_status & ALS_REG_STATUS_DATA_VALID_Msk) {
        result->data_valid = true;
    } else {
        result->data_valid = false;
    }

    if (r_status & ALS_REG_STATUS_DATA_Msk) {
        result->data_new = true;
    } else {
        result->data_new = false;
    }

    uint8_t r_data[2];

    ret = ims_als_ltr_303als_read_register(als, ALS_REG_DATA_CH0_0, &r_data[0]);
    if (ret != IMS_SUCCESS) return ret;

    ret = ims_als_ltr_303als_read_register(als, ALS_REG_DATA_CH0_1, &r_data[1]);
    if (ret != IMS_SUCCESS) return ret;

    result->ch0_value = r_data[1] << 8U | r_data[0];

    ret = ims_als_ltr_303als_read_register(als, ALS_REG_DATA_CH1_0, &r_data[0]);
    if (ret != IMS_SUCCESS) return ret;

    ret = ims_als_ltr_303als_read_register(als, ALS_REG_DATA_CH1_1, &r_data[1]);
    if (ret != IMS_SUCCESS) return ret;

    result->ch1_value = r_data[1] << 8U | r_data[0];

    return IMS_SUCCESS;
}

static ims_ret_t ims_als_ltr_303als_reset(ims_ltr_303als_t *als) {
    uint8_t r_cont;

    if (ims_als_ltr_303als_read_register(als, ALS_REG_CONTR, &r_cont) != IMS_SUCCESS) {
        return IMS_FAIL;
    }

    r_cont |= ALS_REG_CONTR_SW_RST_Msk;

    if (ims_als_ltr_303als_write_register(als, ALS_REG_CONTR, r_cont) != IMS_SUCCESS) {
        return IMS_FAIL;
    }

    do {
        if (ims_als_ltr_303als_read_register(als, ALS_REG_CONTR, &r_cont) != IMS_SUCCESS) {
            return IMS_FAIL;
        }

        als->cb.delay(als->pdev, 5);
    } while (r_cont & ALS_REG_CONTR_SW_RST_Msk);

    return IMS_SUCCESS;
}

static ims_ret_t ims_als_ltr_303als_read_id(ims_ltr_303als_t *als, ims_als_id_t *id) {
    if (ims_als_ltr_303als_read_register(als, ALS_REG_MANUFAC_ID, &id->mfg_id) != IMS_SUCCESS) {
        return IMS_FAIL;
    }

    if (ims_als_ltr_303als_read_register(als, ALS_REG_PART_ID, &id->part_id) != IMS_SUCCESS) {
        return IMS_FAIL;
    }

    return IMS_SUCCESS;
}

static ims_ret_t ims_als_ltr_303als_set_gain(ims_ltr_303als_t *als, ims_ltr_303als_gain_t gain) {
    ims_ret_t ret;

    uint8_t r_cont;

    ret = ims_als_ltr_303als_read_register(als, ALS_REG_CONTR, &r_cont);
    if (ret != IMS_SUCCESS) return ret;

    r_cont &= ~ALS_REG_CONTR_GAIN_Msk;
    r_cont |= ((gain << ALS_REG_CONTR_GAIN_Pos) & ALS_REG_CONTR_GAIN_Msk);

    ret = ims_als_ltr_303als_write_register(als, ALS_REG_CONTR, r_cont);

    return ret;
}

static ims_ret_t ims_als_ltr_303als_set_threshold(ims_ltr_303als_t *als, ims_als_thresh_t type, uint16_t thresh) {
    ims_ret_t ret;

    uint8_t r_thr;

    r_thr = (thresh >> 8U) & 0xFFU;

    ret = ims_als_ltr_303als_write_register(als, type, r_thr);
    if (ret != IMS_SUCCESS) return ret;

    r_thr = thresh & 0xFFU;

    ret = ims_als_ltr_303als_write_register(als, type + 1, r_thr);
    return ret;
}

static ims_ret_t ims_ltr_303als_config_irq(ims_ltr_303als_t *als, ims_ltr_303als_int_polarity_t pol, bool enable) {
    ims_ret_t ret;

    uint8_t r_interrupt;

    ret = ims_als_ltr_303als_read_register(als, ALS_REG_INTERRUPT, &r_interrupt);
    if (ret != IMS_SUCCESS) return ret;

    r_interrupt &= ~(ALS_REG_INTERRUPT_MODE_Msk | ALS_REG_INTERRUPT_POL_Msk);
    r_interrupt |= ((pol << ALS_REG_INTERRUPT_POL_Pos) & ALS_REG_INTERRUPT_POL_Msk);

    if (enable) {
        r_interrupt |= ALS_REG_INTERRUPT_MODE_Msk;
    }

    ret = ims_als_ltr_303als_write_register(als, ALS_REG_INTERRUPT, r_interrupt);

    return ret;
}

static ims_ret_t ims_ltr_303als_config_timing(ims_ltr_303als_t *als, ims_ltr_303als_measurement_rate_t rate,
                                              ims_ltr_303als_integration_time_t time) {
    ims_ret_t ret;

    uint8_t r_measure_rate;

    ret = ims_als_ltr_303als_read_register(als, ALS_REG_MEAS_RATE, &r_measure_rate);
    if (ret != IMS_SUCCESS) return ret;

    r_measure_rate &= ~ALS_REG_MEAS_INT_Msk;
    r_measure_rate |= ((time << ALS_REG_MEAS_INT_Pos) & ALS_REG_MEAS_INT_Msk);

    r_measure_rate &= ~ALS_REG_MEAS_RATE_Msk;
    r_measure_rate |= ((rate << ALS_REG_MEAS_RATE_Pos) & ALS_REG_MEAS_RATE_Msk);

    ret = ims_als_ltr_303als_write_register(als, ALS_REG_MEAS_RATE, r_measure_rate);

    return ret;
}

static ims_ret_t ims_als_ltr_303als_read_register(ims_ltr_303als_t *als, uint8_t reg, uint8_t *value) {
    ims_i2c_xfer_desc_t xfer = {
        .tx_size = 1U,
        .tx_data = &reg,
        .rx_size = 1U,
        .rx_data = value,
    };

    if (als->cb.i2c_xfer(als->pdev, &xfer) != IMS_SUCCESS) {
        return IMS_FAIL;
    }

    return IMS_SUCCESS;
}

static ims_ret_t ims_als_ltr_303als_write_register(ims_ltr_303als_t *als, uint8_t reg, uint8_t value) {
    uint8_t tx_buf[2] = {reg, value};

    ims_i2c_xfer_desc_t xfer = {
        .tx_size = 2U,
        .tx_data = tx_buf,
        .rx_size = 0U,
        .rx_data = NULL,
    };

    if (als->cb.i2c_xfer(als->pdev, &xfer) != IMS_SUCCESS) {
        return IMS_FAIL;
    }

    return IMS_SUCCESS;
}
